function insert_qaqc_campaign_id(db::SQLite.DB)
    campaigns = [1, 2, 3]
    stmt = DBInterface.prepare(db, sql"INSERT INTO qaqc_campaigns VALUES (:id, :note)")
    DBInterface.executemany(stmt, (id = campaigns, note = fill(nothing, size(campaigns))))

    nothing
end

function insert_qaqc_positions(db::SQLite.DB)
    stmt = DBInterface.prepare(
        db,
        sql"""
        INSERT INTO qaqc_positions
        VALUES(
            :id,
            :name,
            :station,
            :position
        )
        """,
    )
    DBInterface.executemany(
        stmt,
        (
            id = 1:18,
            name = ["B-$i-$j" for i in 0:1 for j in 1:9],
            station = [fill(0, 9); fill(1, 9)],
            position = [collect(1:9); collect(1:9)],
        ),
    )

    nothing
end

"""
Common preprocess(format) function for single result table.
"""
function prepare_single_result_df(single_result_table::DataFrame)
    df = copy(single_result_table, copycols = true)

    # timestamp format: 2024-08-07T06:18:09Z
    # ignore the last 'Z' => [1:end-1]
    transform!(
        df,
        :timestamp =>
            ByRow(s -> ismissing(s) ? missing : DateTime(s[1:(end - 1)])) => :timestamp,
    )

    return df
end

"""
Common preprocess(format) function for runlist table.
"""
function prepare_runlist_df(runlist_table::DataFrame)
    df = copy(runlist_table, copycols = true)
end

"""
Add PS Board IDs from single test result table.
"""
function add_psboard_ids(db::SQLite.DB, single_result_table::DataFrame)
    df = combine(groupby(single_result_table, :motherboard_id)) do df
        if df.daughterboard |> unique |> length |> ==(1)
            return (daughterboard = df.daughterboard |> first,)
        end
        df = sort(df, :timestamp)
        dropmissing!(df, :daughterboard)
        return (daughterboard = df.daughterboard |> last,)
    end
    filter!(:motherboard_id => !=(999999), df)

    stmt = DBInterface.prepare(
        db,
        sql"INSERT INTO ps_boards VALUES (:psbid, :daughterboardid)",
    )
    DBInterface.executemany(
        stmt,
        (psbid = df.motherboard_id, daughterboardid = df.daughterboard),
    )

    nothing
end

function add_qaqc_runlist(db::SQLite.DB, runlist_table::DataFrame)
    stmt_insert_runid = DBInterface.prepare(
        db,
        sql"""
        INSERT INTO qaqc_runs
        VALUES (:runid, :campaign_id, :run_datetime, :note, :shifter, :logfile, :shiftscript_ver)
        """,
    )
    runlist_table = dropmissing(runlist_table, Symbol("Run ID"))
    @assert allunique(runlist_table, Symbol("Run ID"))
    for row in eachrow(runlist_table)
        try
            DBInterface.execute(
                stmt_insert_runid,
                (
                    runid = row.var"Run ID",
                    campaign_id = row.var"Campaign ID",
                    run_datetime = nothing,
                    note = row.comment,
                    shifter = row.var"Shifter name",
                    logfile = nothing,
                    shiftscript_ver = nothing,
                ),
            )
        catch e
            @error "error in putting run list" e
            @info "row" row
        end
    end
    nothing
end

function add_qaqc_single_result(
    db::SQLite.DB,
    single_result_table::DataFrame,
    runlist_table::DataFrame,
)
    position_id_map =
        ["B-$i-$j" for i in 0:1 for j in 1:9] |> enumerate .|> (x -> begin
            (i, s) = x
            s => i
        end) |> Dict

    stmt_search_runid = DBInterface.prepare(
        db,
        sql"""
        SELECT id
        FROM qaqc_runs
        WHERE id = :runid
        """,
    )
    stmt_insert_runid = DBInterface.prepare(
        db,
        sql"""
        INSERT INTO qaqc_runs
        VALUES (:runid, :campaign_id, :run_datetime, :note, :shifter, :logfile, :shiftscript_ver)
        """,
    )
    stmt_update_runid = DBInterface.prepare(
        db,
        sql"""
        UPDATE qaqc_runs
        SET run_datetime = :run_datetime, shifter = :shifter, logfile = :logfile, shiftscript_ver = :shiftscript_ver
        WHERE id = :runid
        """,
    )
    stmt_insert_result = DBInterface.prepare(
        db,
        sql"""
        INSERT INTO 
            qaqc_single_run_results(
                runid,
                psboard_id,
                daughterboard_id,
                position,
                resistance_test_passed,
                qspip,
                recov,
                power,
                clock,
                asdtp,
                reset,
                qaqc_result
            )
        VALUES (
                :runid,
                :psboard_id,
                :daughterboard_id,
                :position,
                :resistance_test_passed,
                :qspip,
                :recov,
                :power,
                :clock,
                :asdtp,
                :reset,
                :qaqc_result
            )
        """,
    )
    stmt_insert_resistance = DBInterface.prepare(
        db,
        sql"""
        INSERT INTO qaqc_resistance_check(psb_id, passed)
        VALUES (:psboard_id, :ispassed)
        """,
    )

    for row in eachrow(single_result_table)
        if ismissing(row.runid)
            @assert contains("resistance")(row.comment) || contains("CN15")(row.comment) "Unexpected row with id $(row.motherboard_id) $(row.comment)"

            DBInterface.execute(
                stmt_insert_resistance,
                (psboard_id = row.motherboard_id, ispassed = false),
            )

            continue
        end

        # Add run if it's not in `qaqc_runs` table
        # or update info on the run (such as datetime)
        if DBInterface.execute(stmt_search_runid, (; runid = row.runid)) |> isempty
            campaign_id = if row.runid < 63
                1
            elseif row.runid < 98
                2
            else
                3
            end
            comment = let
                row_run = filter(
                    Symbol("Run ID") => x -> !ismissing(x) && x == row.runid,
                    runlist_table,
                )
                if !isempty(row_run)
                    row_run.comment
                else
                    ""
                end
            end

            DBInterface.execute(
                stmt_insert_runid,
                (
                    runid = row.runid,
                    campaign_id = campaign_id,
                    run_datetime = row.timestamp,
                    note = comment,
                    shifter = row.shifter,
                    logfile = row.qaqc_log_file,
                    shiftscript_ver = row.shiftscript_ver,
                ),
            )
        else
            DBInterface.execute(
                stmt_update_runid,
                (
                    runid = row.runid,
                    run_datetime = row.timestamp,
                    shifter = row.shifter,
                    logfile = row.qaqc_log_file,
                    shiftscript_ver = row.shiftscript_ver,
                ),
            )
        end

        # resistance
        DBInterface.execute(
            stmt_insert_resistance,
            (psboard_id = row.motherboard_id, ispassed = true),
        )

        # main result
        DBInterface.execute(
            stmt_insert_result,
            (
                runid = row.runid,
                psboard_id = row.motherboard_id,
                daughterboard_id = row.daughterboard,
                position = position_id_map[row.position],
                resistance_test_passed = true,
                qspip = row.qspip,
                recov = row.recov,
                power = row.power,
                clock = row.clock,
                asdtp = row.asdtp,
                reset = row.reset,
                qaqc_result = row.qaqc_result,
            ),
        )
    end

    nothing
end

function prepare_dispatch_table(raw_dispatch_table::DataFrame)::DataFrame
    df = copy(raw_dispatch_table, copycols = true)
    transform!(
        df,
        [Symbol("Column$i") for i in 2:ncol(df)] => ByRow((s...) -> join(s |> skipmissing)) => :comment,
    )
    select!(df, [1, ncol(df)])
    rename!(df, [:psboard_id, :comment])

    transform!(
        df,
        :psboard_id => (vs -> accumulate(vs; init = 0) do x, y
            ispsbid = startswith("PS")
            if ispsbid(y)
                x
            else
                match(r"(\d+)", y) |> first
            end
        end) => :campaign_id,
    )
    transform!(
        df,
        :psboard_id =>
            ByRow(s -> startswith("PS")(s) ? parse(Int64, s[3:end]) : missing) =>
                :psboard_id,
    )
    dropmissing!(df, :psboard_id)

    df
end

function add_qaqcdispatch(db::SQLite.DB, dispatch_table::DataFrame)
    dispatch_table = prepare_dispatch_table(dispatch_table)

    # TODO: provide datetime
    stmt = DBInterface.prepare(
        db,
        sql"""
        INSERT INTO qaqc_dispatch(qaqc_campaign_id, psb_id, source_place, destination, time)
        VALUES (:campaign_id, :psboard_id, "KEK", "GND", NULL)
        """,
    )

    DBInterface.executemany(
        stmt,
        (
            campaign_id = dispatch_table.campaign_id,
            psboard_id = dispatch_table.psboard_id,
        )
    )

    nothing
end
